/*
  Copyright 2008, UCAR/Unidata
  See COPYRIGHT file for copying and redistribution conditions.

  This program tests the large file bug in netCDF 3.6.2,
  creating byte and short variables larger than 4 GiB.

  $Id: tst_large.c,v 1.9 2008/04/15 18:30:35 russ Exp $
*/
#include <config.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <nc_tests.h>
#include <netcdf.h>

/* Test with both classic and 64-bit offset files. If netcdf-4 is
 * included, test with both netCDF-4 format variants also. */
#ifdef USE_NETCDF4
#define NUM_FORMATS (4)
#else
#define NUM_FORMATS (2)
#endif

#define NUMDIMS 2		/* rank of each variable in tests */
#define DIM1 2048
#define DIM2 2097153		/* DIM1*DIM2*sizeof(char)   > 2**32 */
#define DIM3 1024
#define DIM4 2097153		/* DIM3*DIM4*sizeof(short)  > 2**32  */

/* 
 * In netCDF-3.6.2, a divide by zero occurs on 32-bit platforms when
 * creating a variable for which the product of dimensions is exactly
 * 2**32.  Check that this bug has been fixed. 
 */
static void
test_big_var(const char *testfile) {
    int ncid, varid, dimids[NUMDIMS];
    int res;
    int cflag = NC_CLOBBER;
    nc_type type = NC_BYTE;
    char *typename = "NC_BYTE";
    int i;
    size_t index[NUMDIMS];
    signed char nval = 99;
    int nval_in;
    
    if ((res = nc_create(testfile, cflag, &ncid)))
	ERR;
    if ((res = nc_set_fill(ncid, NC_NOFILL, NULL)))
	ERR;
    if ((res = nc_def_dim(ncid, "dim1", DIM1, &dimids[0])))
	ERR;
    if ((res = nc_def_dim(ncid, "dim2", DIM2 - 1, &dimids[1])))
	ERR;
    if ((res = nc_def_var(ncid, "var", type, NUMDIMS, dimids, &varid)))
	ERR;
    if ((res = nc_enddef(ncid)))
	ERR;
    index[0] = DIM1 - 1;
    index[1] = DIM2 - 2;
    if ((res = nc_put_var1_schar(ncid, varid, index, &nval)))
	ERR;
    if ((res = nc_close(ncid)))
	ERR;
    if ((res = nc_open(testfile, NC_NOWRITE, &ncid)))
	ERR;
    if ((res = nc_inq_varid(ncid, "var", &varid)))
	ERR;
    if ((res = nc_get_var1_int(ncid, varid, index, &nval_in)))
	ERR;
    if (nval != nval_in)
	ERR;
    if ((res = nc_close(ncid)))
	ERR;
}

static void
test_large_byte_var(const char *testfile) {
    int ncid, varid, dimids[NUMDIMS];
    size_t index[2] = {0, 0};
    int res; 
    int cflag = NC_CLOBBER;
    signed char vals[DIM2];
    signed char char_val_in;
    int i, j;

    if ((res = nc_create(testfile, cflag, &ncid)))
	ERR;
    if ((res = nc_set_fill(ncid, NC_NOFILL, NULL)))
	ERR;
    if ((res = nc_def_dim(ncid, "dim1", DIM1, &dimids[0])))
	ERR;
    if ((res = nc_def_dim(ncid, "dim2", DIM2, &dimids[1])))
	ERR;
    if ((res = nc_def_var(ncid, "var", NC_BYTE, NUMDIMS, dimids, &varid)))
	ERR;
    if ((res = nc_enddef(ncid)))
	ERR;
    for (i=0; i<DIM1; i++) {
	size_t start[NUMDIMS];
	size_t count[NUMDIMS];
	start[0] = i;
	start[1] = 0;
	count[0] = 1;
	count[1] = DIM2;
	for (j=0; j < DIM2; j++) {
	    vals[j] = (i+9)*(j+11); /* note vals[j] is 99 when i==0 and j==0 */
	}
	if ((res = nc_put_vara_schar(ncid, varid, start, count, vals)))
	{
	    ERR;
	    break;
	}
    }
    if ((res = nc_close(ncid)))
	ERR;
    if ((res = nc_open(testfile, NC_NOWRITE, &ncid)))
	ERR;
    if ((res = nc_inq_varid(ncid, "var", &varid)))
	ERR;
    if ((res = nc_get_var1_schar(ncid, varid, index, &char_val_in)))
	ERR;
    if (char_val_in != 99)	/* see above, the value written when i==0, j==0 */
	ERR;
}

static void
test_large_short_var(const char *testfile) {
    int ncid, varid, dimids[NUMDIMS];
    int int_val_in, int_val_out = 99;
    size_t index[2];
    int res; 
    int cflag = NC_CLOBBER;

    if ((res = nc_create(testfile, cflag, &ncid)))
	ERR;
    if ((res = nc_def_dim(ncid, "dim3", DIM3, &dimids[0])))
	ERR;
    if ((res = nc_def_dim(ncid, "dim4", DIM4, &dimids[1])))
	ERR;
    if ((res = nc_def_var(ncid, "var", NC_SHORT, NUMDIMS, dimids, &varid)))
	ERR;
    if ((res = nc_enddef(ncid)))
	ERR;
    index[0] = 0;
    index[1] = 1;
    if ((res = nc_put_var1_int(ncid, varid, index, &int_val_out)))
	ERR;
    if ((res = nc_close(ncid)))
	ERR;
    if ((res = nc_open(testfile, NC_NOWRITE, &ncid)))
	ERR;
    if ((res = nc_inq_varid(ncid, "var", &varid)))
	ERR;
    if ((res = nc_get_var1_int(ncid, varid, index, &int_val_in)))
	ERR;
    if (int_val_in != int_val_out)
	ERR;
#ifndef NOFILL
    index[0] = 1;
    index[1] = 2;
    if ((res = nc_get_var1_int(ncid, varid, index, &int_val_in)))
	ERR;
    if (int_val_in != NC_FILL_SHORT)
	ERR;
#endif
}

#define FILE_NAME "tst_large.nc"

int
main(int argc, char **argv) {
    int i;
    char testfile[NC_MAX_NAME + 1];

    sprintf(testfile, "%s/%s", TEMP_LARGE, FILE_NAME);

    printf("*** testing fix for 3.6.2 large file bug in %s ...\n",
	   testfile);
   /* Go thru formats and run all tests for each of two (for netCDF-3
    * only builds), or 4 (for netCDF-4 builds) different formats. */
/*    for (i = 1; i <= NUM_FORMATS; i++) */
   for (i = NUM_FORMATS; i >= 1; i--)
   {
      switch (i) 
      {
	 case NC_FORMAT_CLASSIC:
	    nc_set_default_format(NC_FORMAT_CLASSIC, NULL);
	    printf("Switching to netCDF classic format.\n");
	    break;
	 case NC_FORMAT_64BIT:
	    nc_set_default_format(NC_FORMAT_64BIT, NULL);
	    printf("Switching to 64-bit offset format.\n");
	    break;
#ifdef USE_NETCDF4
	 case NC_FORMAT_NETCDF4_CLASSIC:
	    nc_set_default_format(NC_FORMAT_NETCDF4_CLASSIC, NULL);
	    printf("Switching to netCDF-4 format (with NC_CLASSIC_MODEL).\n");
	    break;
	 case NC_FORMAT_NETCDF4: /* actually it's _CLASSIC. */
	    nc_set_default_format(NC_FORMAT_NETCDF4, NULL);
	    printf("Switching to netCDF-4 format.\n");
	    break;
#endif
	 default:
	    printf("Unexpected format!\n");
	    return 2;
      }
      
      printf("*** testing creating a variable with 2**32 values...");
      test_big_var(testfile);
      (void) remove(testfile);
      SUMMARIZE_ERR;
      
      printf("*** testing accessing byte variable with > 2**32 values...");
      test_large_byte_var(testfile);
      (void) remove(testfile);
      SUMMARIZE_ERR;

      printf("*** testing accessing short variable with > 2**32 values...");
      test_large_short_var(testfile);
      (void) remove(testfile);
      SUMMARIZE_ERR;
   }

   FINAL_RESULTS;
}
